;
; Copyright (C) 2006 Johannes Hausensteiner (johannes.hausensteiner@pcl.at)
;
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
;
;
; Filename: diy_calc.asm
;
; Function: ROM monitor for the DIY Calculator
;
;
; assemble with tasm with following parameters:
;   tasm -01 -i -o10 -c -f0 diy_calc.asm diy_calc.hex
;
; generate the .mem file for import into ispLEVER FPGA project
;   hex2mem diy_calc
;
;
; Changelog
;
;  0.1    6.Sep.2006   JH   new
;  0.2   13.Sep.2006   JH   RAM file loader, add reset cmd
;  0.3   13.Sep.2006   JH   solved error in atox(), atolx()
;                           UART interrupt routine, ring buffer
;                           CPU clock control
;                           HW handshake first version; does not yet work
;

#include "diy_calc.inc"
#include "monitor.inc"

VERSION:   .EQU    $03

USE_RX_IRQ .EQU    1

MON_RAM    .EQU    $FE00          ; special RAM area for the ROM monitor

CMD_HELP   .EQU    $00            ; display help
CMD_VER    .EQU    $01            ; display version information
CMD_MRD    .EQU    $02            ; memory read cmd
CMD_MWR    .EQU    $03            ; memory write cmd
CMD_LOAD   .EQU    $04            ; load a program
CMD_RUN    .EQU    $05            ; run the program
CMD_RST    .EQU    $06            ; do a warm start
CMD_RDMP   .EQU    $07            ; CPU register dump
CMD_TEST   .EQU    $08            ; for tests

CTRL_C     .EQU    $03            ; CTRL C
BEL        .EQU    $07            ; bell
BS         .EQU    $08            ; backspace
LF         .EQU    $0A            ; line feed
CR         .EQU    $0D            ; carriage return
CTRL_Q     .EQU    $11            ; CTRL Q
CTRL_S     .EQU    $13            ; CTRL S

           .ORG    MON_RAM
PUTC       .DW                    ; pointer to putc() function
PUTS       .DW                    ; pointer to puts() function
TSTC       .DW                    ; pointer ot tstc() function
GETC       .DW                    ; pointer to getc() function
GETCHE     .DW                    ; pointer to getche() function
GETCHAR    .DW                    ; pointer to getchar() function
GETS       .DW                    ; pointer to gets() function

; global use
PTR16:     .DW                    ; 16bit pointer
TMP16:     .DW                    ; temp 16bit memory
TMP:       .DB                    ; temp 8bit memory
ERROR:     .DB                    ; return value of functions
BUF        .BLOCK  80             ; general purpose buffer
CMD_OPT:   .DW                    ; ptr to current cmdline option
TMP_ACC:   .DB
TMP_SR:    .DB
TMP_IR:    .DB
TMP_PC:    .DW
TMP_X:     .DW
TMP_SP:    .DW
TMP_IV:    .DW
RX_STOP:   .DB                    ; for RS-232 hardware handshake
TX_STOP:   .DB                    ; for RS-232 hardware handshake

; use in main()
CMDLINE:   .BLOCK  80             ; cmdline
CMDSTR:    .BLOCK  16             ; the cmd extracted from cmdline
CMD:       .DB                    ; enumerated cmd value

; use in strcpy(), strcmp(), strchr()
R_ADD:     .DW                    ; return address
S1:        .DW                    ; destination string
S2:        .DW                    ; source string

; use in atox(), atolx()
NUMBER:    .DW                    ; the accumulated 16bit number
NEGATIVE:  .DB                    ; flag for neagtive
DIGITS:    .DB                    ; current number of digit
MAXDIGIT:  .DB                    ; max number of digits

; use in md()
MD_START:  .DW                    ; start address of memory region (md cmd)
MD_END:    .DW                    ; end address of memory region (md cmd)
MD_ADDR:   .DW                    ; current address of memory region (md cmd)
MD_NUM:    .DW                    ; number of bytes to write/display (md cmd)

; use in mw()
MW_START:  .DW                    ; start address of memory region (mw cmd)
MW_END:    .DW                    ; end address of memory region (mw cmd)
MW_ADDR:   .DW                    ; current address of memory region (mw cmd)
MW_NUM:    .DW                    ; number of bytes to write/display (mw cmd)
MW_DATA:   .DB                    ; write data

; use in load()
ADDR:      .DW                    ; current address
END_ADDR:  .DW                    ; end address

; use in interrupt handler
IRQ_ACC:   .DB                    ; save acc
IRQ_X:     .DW                    ; save X
IRQ_ISRC:  .DB                    ; save IRQ source port

; ring buffer for UART Rx interrupt, must be aligned to RX_BSIZE
RX_RPTR:   .DW                    ; ring buffer read pointer
RX_WPTR:   .DW                    ; ring buffer write pointer
RX_BMAX:   .DB                    ; max. ringbuffer addres, low byte only
RX_BSIZE:  .EQU    64             ; buffer size
           .ORG    (($ + RX_BSIZE) & (-RX_BSIZE))
RX_BUF:    .BLOCK  RX_BSIZE       ; the ring buffer itself
RX_THRES:  .EQU    RX_BSIZE/4     ; number of bytes left before we stop the
                                  ; device on the other end of the RS232 line

           .ORG    $0000
RESET:     BLDSP   RAMEND
           LDA     BEEPCODE
           STA     [MAINDISP]
           LDA     0
           STA     [CMD]
           STA     [RX_STOP]
           STA     [TX_STOP]
; set maximum cpu clock
           STA     [CPU_CLK]
; check for warm/cold reset           
           LDA     [IRQ_SRC]
           AND     IRQ_HARD
           JZ      [MAIN]
INIT:
; enable "automatic" RTS signal
           LDA     $90
           STA     [UART_CTL]
; print welcome message     
           LDA     CR
           JSR     [_PUTC]
           LDA     LF
           JSR     [_PUTC]
           BLDX    MSG001
           JSR     [_PUTS]
; initialize global variables for md, mw commands
           BLDX    $4000
           BSTX    [MD_START]
           BLDX    $0100
           BSTX    [MD_NUM]
           BLDX    $4000
           BSTX    [MW_START]
           BLDX    1
           BSTX    [MW_NUM]
; clear LCD and write welcome message
           LDA     CLRCODE
           JSR     [LCDPUT]
           BLDX    MSG001
           JSR     [LCDPUTS]
           LDA     0
           STA     [SIXLEDS]
; assign function pointers
           BLDX    _PUTC
           BSTX    [PUTC]
           BLDX    _PUTS
           BSTX    [PUTS]
           BLDX    _TSTC
           BSTX    [TSTC]
           BLDX    _GETC
           BSTX    [GETC]
           BLDX    _GETCHE
           BSTX    [GETCHE]
           BLDX    _GETCHAR
           BSTX    [GETCHAR]
           BLDX    _GETS
           BSTX    [GETS]
; compute end address of ring buffer
           BLDX    RX_BUF
           BSTX    [TMP16]
           LDA     [TMP16+1]
           ADD     RX_BSIZE-1
           STA     [RX_BMAX]
; initialize ring buffer write and read pointers
           BLDX    RX_BUF
           BSTX    [RX_WPTR]
           BSTX    [RX_RPTR]

; setup interrupt handler and enable interrupts
           BLDIV   IRQ_HAND
           BSTIV   [TMP_IV]
#if USE_RX_IRQ
           SETIM
; enable Rx interrupt from UART
           LDA     [UART_STA]
           OR      UART_RXIE
           STA     [UART_CTL]
#endif
; print welcome message
           BLDX    MSG002
           JSR     [puts]
           LDA     VERSION
           JSR     [PUTVER]
           JSR     [newline]
           LDA     BEEPCODE
           JSR     [LCDPUT]

; mainloop
MAIN:      JSR     [PROMPT]
; read the command
           BLDX    CMDLINE
           JSR     [gets]

; extract the command from the cmdline
           BLDX    CMDLINE
           BSTX    [TMP16]
           BLDX    CMDSTR
           BSTX    [PTR16]
           BLDX    0
LBL004:    LDA     [[TMP16], X]
           STA     [[PTR16], X]
           JZ      [LBL003]       ; no cmdline option
           INCX
           CMPA    ' '
           JNZ     [LBL004]
           DECX
LBL003:    INCX
           DECX
           JZ      [LBL002]       ; cmdline was empty => exec previous cmd
           LDA     0              ; 0-terminate the command
           STA     [[PTR16], X]

; analyze the command
           BLDX    CMDSTR
           PUSHX
           BLDX    CMDSTR_HELP
           PUSHX
           JSR     [STRCMP]
           JNZ     [LBL005]
           LDA     CMD_HELP       ; "help" cmd
           JMP     [LBL001]

LBL005:    BLDX    CMDSTR
           PUSHX
           BLDX    CMDSTR_VER
           PUSHX
           JSR     [STRCMP]
           JNZ     [LBL006]
           LDA     CMD_VER        ; "ver" cmd
           JMP     [LBL001]

LBL006:    BLDX    CMDSTR
           PUSHX
           BLDX    CMDSTR_MRD
           PUSHX
           JSR     [STRCMP]
           JNZ     [LBL007]
           BLDX    CMDLINE
           LDA     CMD_MRD        ; "md" cmd
           JMP     [LBL001]

LBL007:    BLDX    CMDSTR
           PUSHX
           BLDX    CMDSTR_MWR
           PUSHX
           JSR     [STRCMP]
           JNZ     [LBL008]
           BLDX    CMDLINE
           LDA     CMD_MWR        ; "mw" cmd
           JMP     [LBL001]

LBL008:    BLDX    CMDSTR
           PUSHX
           BLDX    CMDSTR_LOAD
           PUSHX
           JSR     [STRCMP]
           JNZ     [LBL009]
           LDA     CMD_LOAD       ; "load" cmd
           JMP     [LBL001]

LBL009:    BLDX    CMDSTR
           PUSHX
           BLDX    CMDSTR_RUN
           PUSHX
           JSR     [STRCMP]
           JNZ     [LBL010]
           LDA     CMD_RUN        ; "run" cmd
           JMP     [LBL001]

LBL010:    BLDX    CMDSTR
           PUSHX
           BLDX    CMDSTR_RST
           PUSHX
           JSR     [STRCMP]
           JNZ     [LBL011]
           LDA     CMD_RST        ; "reset" cmd
           JMP     [LBL001]

LBL011:    BLDX    CMDSTR
           PUSHX
           BLDX    CMDSTR_RDMP
           PUSHX
           JSR     [STRCMP]
           JNZ     [LBL012]
           LDA     CMD_RDMP       ; "rdump" cmd
           JMP     [LBL001]

LBL012:    BLDX    CMDSTR
           PUSHX
           BLDX    CMDSTR_TEST
           PUSHX
           JSR     [STRCMP]
           JNZ     [LBL013]
           LDA     CMD_TEST       ; "test" cmd
           JMP     [LBL001]

LBL013:    BLDX    CMDSTR_UNKNOWN1
           JSR     [puts]
           BLDX    CMDSTR
           JSR     [puts]
           BLDX    CMDSTR_UNKNOWN2
           JSR     [puts]
           LDA     0
           STA     [CMD]
           JMP     [MAIN]

LBL001:    STA     [CMD]

; build jump table
LBL002:    LDA     0
           STA     [TMP16]
           LDA     [CMD]
           SHL
           SHL
           STA     [TMP16+1]
           BLDX    [TMP16]
           JSR     [CMD_TABLE, X]
           JMP     [MAIN]

; cmd jump table, each entry must be 4 bytes
CMD_TABLE: JMP     [HELP_CMD]
           NOP

           JMP     [VER_CMD]
           NOP

           JMP     [MD_CMD]
           NOP

           JMP     [MW_CMD]
           NOP

           JMP     [LOAD_CMD]
           NOP

           JMP     [RUN_CMD]
           NOP

           JMP     [RST_CMD]
           NOP

           JMP     [RDMP_CMD]
           NOP

           JMP     [TEST_CMD]
           NOP

CMDSTR_HELP: .BYTE "help", 0      ; display help
CMDSTR_VER:  .BYTE "ver", 0       ; display version information
CMDSTR_MRD:  .BYTE "md", 0        ; memory display cmd
CMDSTR_MWR:  .BYTE "mw", 0        ; memory write cmd
CMDSTR_LOAD: .BYTE "load", 0      ; load program cmd
CMDSTR_RUN:  .BYTE "run", 0       ; run program cmd
CMDSTR_RST:  .BYTE "reset", 0     ; do a warm start
CMDSTR_RDMP: .BYTE "rdump", 0     ; CPU register dump
CMDSTR_TEST: .BYTE "test", 0      ; for test
CMDSTR_UNKNOWN1: .BYTE "Unknown command \"", 0
CMDSTR_UNKNOWN2: .BYTE "\" - try \"help\"", CR, LF, 0

MSG001:    .BYTE   "DIY Calculator ", 0
MSG002:    .BYTE   "ROM monitor V", 0
MSG003:    .BYTE   "CPU Rev: ", 0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function putver: print the ROM monitor version in x.y format
;
; Input:  A .. version byte
;
; Output: --
;
; Uses:   A, C, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
PUTVER:
; process high nibble
           SHR
           SHR
           SHR
           SHR
           AND     $0F
           CMPA    9
           JC      [PUTVER1]
           ADD     '0'
           JMP     [PUTVER2]
PUTVER1:   ADD     'A'-10
PUTVER2:   JSR     [putchar]
           LDA     '.'
           JSR     [putchar]
; process low nibble
           LDA     VERSION
           AND     $0F
           CMPA    9
           JC      [PUTVER3]
           ADD     '0'
           JMP     [PUTVER4]
PUTVER3:   ADD     'A'-10
PUTVER4:   JSR     [putchar]
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function help_cmd: execute the help cmd
;
; Input:  --
;
; Output: --
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
HELP_CMD:
; call getarg() .. argv[0]
           BLDX    BUF
           PUSHX
           BLDX    CMDLINE
           PUSHX
           JSR     [GETARG]
           POPX
           BSTX    [CMD_OPT]
           LDA     [[CMD_OPT]]
           JZ      [HELP_CMD2]    ; no arg

; call getarg() .. argv[1]
           BLDX    BUF
           PUSHX
           BLDX    [CMD_OPT]
           PUSHX
           JSR     [GETARG]

; get ptr to the "new cmdline"
           POPX
           BSTX    [CMD_OPT]
                                    
; decode cmdline arg
           BLDX    BUF
           PUSHX
           BLDX    CMDSTR_MRD
           PUSHX
           JSR     [STRCMP]
           JZ      [HELP_MRD]

           BLDX    BUF
           PUSHX
           BLDX    CMDSTR_MWR
           PUSHX
           JSR     [STRCMP]
           JZ      [HELP_MWR]

; no extended help for this cmd available
           BLDX    HELP_MS11
           JSR     [puts]
           JMP     [HELP_CMD1]

HELP_MRD:  BLDX    HELP_MS12
           JSR     [puts]
           JMP     [HELP_CMD1]

HELP_MWR:  BLDX    HELP_MS13
           JSR     [puts]
           JMP     [HELP_CMD1]

; no args given
HELP_CMD2: BLDX    HELP_MSG1
           JSR     [puts]
           BLDX    HELP_MSG2
           JSR     [puts]
           BLDX    HELP_MSG3
           JSR     [puts]
           BLDX    HELP_MSG4
           JSR     [puts]
           BLDX    HELP_MSG5
           JSR     [puts]
           BLDX    HELP_MSG6
           JSR     [puts]
           BLDX    HELP_MSG7
           JSR     [puts]
           BLDX    HELP_MSG8
           JSR     [puts]
           BLDX    HELP_MSG9
           JSR     [puts]
           BLDX    HELP_MS10
           JSR     [puts]
           JMP     [HELP_CMD1]

HELP_CMD1: RTS

HELP_MSG1: .BYTE   CR, LF, "Following commands are available:", CR, LF, 0
HELP_MSG2: .BYTE   "  help ... this text", CR, LF, 0
HELP_MSG3: .BYTE   "  ver .... display version information", CR, LF, 0
HELP_MSG4: .BYTE   "  md ..... memory display", CR, LF, 0
HELP_MSG5: .BYTE   "  mw ..... memory write", CR, LF, 0
HELP_MSG6: .BYTE   "  load ... load a program (RAM format)", CR, LF, 0
HELP_MSG7: .BYTE   "  run .... run the program at address $4000", CR, LF, 0
HELP_MSG8: .BYTE   "  reset .. perform a soft reset", CR, LF, 0
HELP_MSG9: .BYTE   "  rdump .. CPU register dump", CR, LF, 0
HELP_MS10: .BYTE   CR, LF, "For extended help type 'help cmd'", CR, LF, 0
HELP_MS11: .BYTE   CR,LF,"No extended help for this command available",CR,LF,0
HELP_MS12: .BYTE   CR,LF,"md [address] [no. of bytes]", CR, LF
           .BYTE   "   address is auto-incremented", CR, LF, 0
HELP_MS13: .BYTE   CR,LF,"mw data [address] [no. of bytes]", CR, LF, 0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function ver_cmd: execute the version cmd
;
; Input:  --
;
; Output: --
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
VER_CMD:
; FPGA version
           BLDX    VER_MSG1
           JSR     [puts]
           LDA     [HW_REV]
           JSR     [PUTHEX]
; CPU version
           BLDX    VER_MSG2
           JSR     [puts]
           VER
           JSR     [PUTHEX]
; ROM monitor version
           BLDX    VER_MSG3
           JSR     [puts]
           LDA     VERSION
           JSR     [PUTVER]
; CPU clock divider
           BLDX    VER_MSG4
           JSR     [puts]
           JSR     [CPU_DIV]
           PUSHA
           JSR     [PUTHEX16]
           POPA
           JSR     [PUTHEX]

           JSR     [newline]
           RTS

VER_MSG1:  .BYTE   CR, LF, "FPGA version: $", 0
VER_MSG2:  .BYTE   CR, LF, "CPU  version: $", 0
VER_MSG3:  .BYTE   CR, LF, "ROM monitor version: V", 0
VER_MSG4:  .BYTE   CR, LF, "CPU clock divider: $", 0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function CPU_DIV: compute the CPU clock divider
;
; Input:  --
;
; Output: A .. low byte
;         X .. two high bytes
;
; Uses:   A, X
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CPU_DIV:
           LDA     [CPU_CLK]
           SHR
           SHR
           SHR
           SHR
           AND     $0F
           JZ      [CPU_DIV0]
           CMPA    1
           JZ      [CPU_DIV1]
           CMPA    2
           JZ      [CPU_DIV2]
           CMPA    3       
           JZ      [CPU_DIV3]
           CMPA    4
           JZ      [CPU_DIV4]
           CMPA    5
           JZ      [CPU_DIV5]
           CMPA    6
           JZ      [CPU_DIV6]
           CMPA    7
           JZ      [CPU_DIV7]
           CMPA    8
           JZ      [CPU_DIV8]
           CMPA    9
           JZ      [CPU_DIV9]
           CMPA    10
           JZ      [CPU_DIV10]
           CMPA    11
           JZ      [CPU_DIV11]
           CMPA    12
           JZ      [CPU_DIV12]
           CMPA    13
           JZ      [CPU_DIV13]
           CMPA    14
           JZ      [CPU_DIV14]

;15: 1000 0000 0000 000x xxx0 0000
           BLDX    $8000
           LDA     [CPU_CLK]
           SHL
           SHL
           SHL
           SHL
           JNN     [CPU_D151]
           INCX
CPU_D151:  SHL
           JMP     [CPU_DIV15]

CPU_DIV0:
; 0000 0000 0000 0000 000x xxx0
           BLDX    $0000
           LDA     [CPU_CLK]
           AND     $0F
           SHL
           JMP     [CPU_DIV15]

CPU_DIV1:
; 0000 0000 0000 0000 001x xxx0
           BLDX    $0000
           LDA     [CPU_CLK]
           AND     $0F
           SHL
           OR      $20
           JMP     [CPU_DIV15]

CPU_DIV2:
; 0000 0000 0000 0000 010x xxx0
           BLDX    $0000
           LDA     [CPU_CLK]
           AND     $0F
           SHL
           OR      $40
           JMP     [CPU_DIV15]

CPU_DIV3:
; 0000 0000 0000 0000 100x xxx0
           BLDX    $0000
           LDA     [CPU_CLK]
           AND     $0F
           SHL
           OR      $80
           JMP     [CPU_DIV15]

CPU_DIV4:
; 0000 0000 0000 0010 000x xxx0
           BLDX    $0002
           LDA     [CPU_CLK]
           AND     $0F
           SHL
           JMP     [CPU_DIV15]

CPU_DIV5:
; 0000 0000 0000 0100 00xx xx00
           BLDX    $0004
           LDA     [CPU_CLK]
           AND     $0F
           SHL
           SHL
           JMP     [CPU_DIV15]

CPU_DIV6:
; 0000 0000 0001 0000 00xx xx00
           BLDX    $0010
           LDA     [CPU_CLK]
           AND     $0F
           SHL
           SHL
           JMP     [CPU_DIV15]

CPU_DIV7:
; 0000 0000 0010 0000 00xx xx00
           BLDX    $0020
           LDA     [CPU_CLK]
           AND     $0F
           SHL
           SHL
           JMP     [CPU_DIV15]

CPU_DIV8:
; 0000 0000 0100 0000 0xxx x000
           BLDX    $0040
           LDA     [CPU_CLK]
           AND     $0F
           SHL
           SHL
           SHL
           JMP     [CPU_DIV15]

CPU_DIV9:
; 0000 0001 0000 0000 0xxx x000
           BLDX    $0100
           LDA     [CPU_CLK]
           AND     $0F
           SHL
           SHL
           SHL
           JMP     [CPU_DIV15]

CPU_DIV10:
; 0000 0010 0000 0000 0xxx x000
           BLDX    $0200
           LDA     [CPU_CLK]
           AND     $0F
           SHL
           SHL
           SHL
           JMP     [CPU_DIV15]

CPU_DIV11:
; 0000 0100 0000 0000 xxxx 0000
           BLDX    $0400
           LDA     [CPU_CLK]
           SHL
           SHL
           SHL
           SHL
           JMP     [CPU_DIV15]

CPU_DIV12:
; 0001 0000 0000 0000 xxxx 0000
           BLDX    $1000
           LDA     [CPU_CLK]
           SHL
           SHL
           SHL
           SHL
           JMP     [CPU_DIV15]

CPU_DIV13:
; 0010 0000 0000 0000 xxxx 0000
           BLDX    $2000
           LDA     [CPU_CLK]
           SHL
           SHL
           SHL
           SHL
           JMP     [CPU_DIV15]

CPU_DIV14:
; 0100 0000 0000 000x xxx0 0000
           BLDX    $4000
           LDA     [CPU_CLK]
           SHL
           SHL
           SHL
           SHL
           JNN     [CPU_D141]
           INCX
CPU_D141:  SHL

CPU_DIV15:
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function md_cmd: execute the memory display cmd:
;                  md address [# of bytes]
;
; Input:  X ... ptr to cmdline
;
; Output: X .. next address
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
MD_CMD:    BLDX    MD_MSG1
           JSR     [puts]

; extract cmdline parameters
           BLDX    CMDLINE
           BSTX    [CMD_OPT]
; skip the cmd itself
MD_CMD1:   LDA     [[CMD_OPT]]
           JZ      [MD_CMD2]      ; end of string, no parameters
           BLDX    [CMD_OPT]
           INCX
           BSTX    [CMD_OPT]
           CMPA    ' '
           JNZ     [MD_CMD1]

; call strcpy()
           BLDX    BUF
           PUSHX
           BLDX    [CMD_OPT]
           PUSHX
           JSR     [STRCPY]

; convert string to number
           BLDX    BUF
           JSR     [ATOLX]
           BSTX    [MD_START]

; get the number of bytes
MD_CMD3:   LDA     [[CMD_OPT]]
           JZ      [MD_CMD2]      ; end of string, no parameters
           BLDX    [CMD_OPT]
           INCX
           BSTX    [CMD_OPT]
           CMPA    ' '
           JNZ     [MD_CMD3]

; call strcpy()
           BLDX    BUF
           PUSHX
           BLDX    [CMD_OPT]
           PUSHX
           JSR     [STRCPY]

; convert string to number
           BLDX    BUF
           JSR     [ATOLX]
           BSTX    [MD_NUM]

; compute end address
MD_CMD2:   LDA     [MD_START+1]
           ADD     [MD_NUM+1]
           STA     [MD_END+1]
           LDA     [MD_START]
           ADDC    [MD_NUM]
           STA     [MD_END]

; print start + end address
           LDA     [MD_START]
           JSR     [PUTHEX]
           LDA     [MD_START+1]
           JSR     [PUTHEX]
           BLDX    MD_MSG2
           JSR     [puts]
           BLDX    [MD_END]
           DECX
           BSTX    [PTR16]
           LDA     [PTR16]
           JSR     [PUTHEX]
           LDA     [PTR16+1]
           JSR     [PUTHEX]
           JSR     [newline]

; display address
MD_CMD6:   BLDX    [MD_START]
           BSTX    [MD_ADDR]
           BLDX    0
           LDA     [MD_START]
           JSR     [PUTHEX]
           LDA     [MD_START+1]
           JSR     [PUTHEX]

           LDA     ' '
           JSR     [putchar]

; display data
MD_CMD7:   LDA     ' '
           JSR     [putchar]
           LDA     [[MD_ADDR]]
           JSR     [PUTHEX]
; increment and check if last
           INCX
           BSTX    [TMP16]
           BLDX    [MD_ADDR]
           INCX
           BSTX    [MD_ADDR]
; compare high byte
           LDA     [MD_ADDR]
           CMPA    [MD_END]
           JC      [MD_CMD9]      ; current addr > end addr => finished
           JNZ     [MD_CMD8]      ; current < end => continue
; compare low byte
           LDA     [MD_ADDR+1]
           CMPA    [MD_END+1]
           JC      [MD_CMD9]      ; start > end => finished
           JZ      [MD_CMD9]      ; start == end => finished

; line done after 16 bytes
MD_CMD8:   LDA     [TMP16+1]
           AND     $0F
           JZ      [MD_CMD11]     ; done with HEX => continue with ASCII

; insert additional blank after 8 bytes
           AND     $07
           JNZ     [MD_CMD7]      ; next byte
           LDA     ' '
           JSR     [putchar]
           JMP     [MD_CMD7]      ; next byte

; we reached the end of display area => fill HEX line with ' '
MD_CMD9:
; have we reached the end of line (16 HEX bytes) already?
           LDA     [TMP16+1]
           AND     $0F
           JZ      [MD_CMD11]     ; yes => continue with ASCII
; insert additional blank after 8 bytes
           AND     $07
           JNZ     [MD_CMD17]
           LDA     ' '
           JSR     [putchar]
MD_CMD17:  LDA     ' '
           JSR     [putchar]      ; one for the ' '
           LDA     ' '
           JSR     [putchar]      ; one for the high nibble
           LDA     ' '
           JSR     [putchar]      ; one for the low nibble
           INCX
           BSTX    [TMP16]
           JMP     [MD_CMD9]      ; loop until 16 "bytes" filled

; display ASCII
MD_CMD11:  LDA     ' '
           JSR     [putchar]
           LDA     ' '
           JSR     [putchar]
           LDA     ' '
           JSR     [putchar]
           BLDX    [MD_START]
           BSTX    [MD_ADDR]
           BLDX    0
MD_CMD12:  LDA     [[MD_ADDR]]
           CMPA    $20
           JC      [MD_CMD13]
           LDA     '.'            ; replace non-printable characters with '.'
MD_CMD13:  CMPA    $7E
           JNC     [MD_CMD15]
           LDA     '.'
MD_CMD15:  JSR     [putchar]
; increment and check if last
           INCX
           BSTX    [TMP16]
           BLDX    [MD_ADDR]
           INCX
           BSTX    [MD_ADDR]
; compare high byte
           LDA     [MD_ADDR]
           CMPA    [MD_END]
           JC      [MD_CMD16]     ; current > end => finished
           JNZ     [MD_CMD18]
; compare low byte
           LDA     [MD_ADDR+1]
           CMPA    [MD_END+1]
           JC      [MD_CMD16]     ; start > end => finished
           JZ      [MD_CMD16]     ; start == end => finished

; line done after 16 characters
MD_CMD18:  LDA     [TMP16+1]
           AND     $0F
           JZ      [MD_CMD14]     ; line is done (HEX + ASCII)

; insert additional blank after 8 characters
           AND     $07
           JNZ     [MD_CMD12]     ; next byte
           LDA     ' '
           JSR     [putchar]
           JMP     [MD_CMD12]     ; next byte

MD_CMD14:  LDA     CR
           JSR     [putchar]
           LDA     LF
           JSR     [putchar]
; update start address
           BLDX    [MD_ADDR]
           BSTX    [MD_START]
           JMP     [MD_CMD6]      ; next line

; we are done, save address for next run
MD_CMD16:
; update start address
           BLDX    [MD_ADDR]
           BSTX    [MD_START]
           JSR     [newline]
           RTS

MD_MSG1:   .BYTE   CR, LF, "Memory display from $", 0
MD_MSG2:   .BYTE   " to $", 0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function mw_cmd: execute the memory write cmd
;
; Input:  --
;
; Output: --
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
MW_CMD:    BLDX    MW_MSG1
           JSR     [puts]

; extract cmdline parameters
           BLDX    CMDLINE
           BSTX    [CMD_OPT]
; skip the cmd itself
MW_CMD1:   LDA     [[CMD_OPT]]
           JZ      [MW_CMD2]      ; end of string, no parameters
           BLDX    [CMD_OPT]
           INCX
           BSTX    [CMD_OPT]
           CMPA    ' '
           JNZ     [MW_CMD1]

; call strcpy()  .. the write data
           BLDX    BUF
           PUSHX
           BLDX    [CMD_OPT]
           PUSHX
           JSR     [STRCPY]

; convert string to number
           BLDX    BUF
           JSR     [ATOX]
           STA     [MW_DATA]

; get the next arq
MW_CMD3:   LDA     [[CMD_OPT]]
           JZ      [MW_CMD2]      ; end of string, no more parameters
           BLDX    [CMD_OPT]
           INCX
           BSTX    [CMD_OPT]
           CMPA    ' '
           JNZ     [MW_CMD3]

; call strcpy()  .. the start address
           BLDX    BUF
           PUSHX
           BLDX    [CMD_OPT]
           PUSHX
           JSR     [STRCPY]

; convert string to number
           BLDX    BUF
           JSR     [ATOLX]
           BSTX    [MW_START]

; get the next arg
MW_CMD4:   LDA     [[CMD_OPT]]
           JZ      [MW_CMD2]      ; end of string, no more parameters
           BLDX    [CMD_OPT]
           INCX
           BSTX    [CMD_OPT]
           CMPA    ' '
           JNZ     [MW_CMD4]

; call strcpy()  .. the number of bytes
           BLDX    BUF
           PUSHX
           BLDX    [CMD_OPT]
           PUSHX
           JSR     [STRCPY]

; convert string to number
           BLDX    BUF
           JSR     [ATOLX]
           BSTX    [MW_NUM]

MW_CMD2:   BLDX    [MW_NUM]
           DECX
           JZ      [MW_CMD10]     ; only one byte to write
           LDA     'e'
           JSR     [putchar]
           LDA     's'
           JSR     [putchar]
MW_CMD10:  LDA     ' '
           JSR     [putchar]
           LDA     '$'
           JSR     [putchar]

; compute end address
           LDA     [MW_START+1]
           ADD     [MW_NUM+1]
           STA     [MW_END+1]
           LDA     [MW_START]
           ADDC    [MW_NUM]
           STA     [MW_END]

; print start + end address
           LDA     [MW_START]
           JSR     [PUTHEX]
           LDA     [MW_START+1]
           JSR     [PUTHEX]
           BLDX    [MW_NUM]
           DECX
           JZ      [MW_CMD6]      ; only 1 byte to write
           BLDX    MW_MSG2
           JSR     [puts]
           BLDX    [MW_END]
           DECX
           BSTX    [PTR16]
           LDA     [PTR16]
           JSR     [PUTHEX]
           LDA     [PTR16+1]
           JSR     [PUTHEX]
MW_CMD6:   BLDX    MW_MSG3
           JSR     [puts]
           LDA     [MW_DATA]
           JSR     [PUTHEX]
           JSR     [newline]

; now do stuff
           BLDX    0
MW_CMD7:   LDA     [MW_DATA]
           STA     [[MW_START], X]
; increment and compare
           INCX
           BSTX    [TMP16]
; compare high byte
           LDA     [MW_NUM]
           CMPA    [TMP16]
           JC      [MW_CMD7]      ; not yet done
           JZ      [MW_CMD8]      ; equal => compare low byte
           JMP     [MW_CMD9]      ; done

MW_CMD8:   LDA     [MW_NUM+1]
           CMPA    [TMP16+1]
           JC      [MW_CMD7]      ; not yet done

MW_CMD9:   RTS

MW_MSG1:   .BYTE   CR, "Writing address", 0
MW_MSG2:   .BYTE   " to $", 0
MW_MSG3:   .BYTE   " with $", 0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function load_cmd: execute the load cmd
;
; Input:  --
;
; Output: --
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
LOAD_CMD:  BLDX    LOAD_MSG1
           JSR     [puts]

           JSR     [LOADRAM]
           JNZ     [LD_CMD1]

           BLDX    LOAD_MSG2
           JSR     [puts]
           JMP     [LD_CMD2]

LD_CMD1:   BLDX    LOAD_MSG3
           JSR     [puts]
LD_CMD2:   RTS

LOAD_MSG1: .BYTE   CR, LF, "Waiting for RAM file ... ", CR, LF, 0
LOAD_MSG2: .BYTE   CR, LF, "Success!", CR, LF, 0
LOAD_MSG3: .BYTE   CR, LF, "Error!", CR, LF, 7, 0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function run_cmd: execute the run cmd
;
; Input:  --
;
; Output: --
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
RUN_CMD:   BLDX    RUN_MSG1
           JSR     [puts]
; since we were called as subroutine clean the stack
           POPA
           POPA
           JMP     [RAMSTART]

RUN_MSG1:  .BYTE   CR, LF, "Executing JMP [$4000] ... ", 0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function rst_cmd: do a warm start
;
; Input:  --
;
; Output: --
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
RST_CMD:   BLDX    RST_MSG1
           JSR     [puts]
           JMP     [RESET]        ; stack will be initialized there

RST_MSG1:  .BYTE   CR, LF, "rebooting ... ", CR, LF, 0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function rdmp_cmd: CPU register dump
;
; Input:  --
;
; Output: --
;
; Uses:   A, X
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
RDMP_CMD:
; save all registers
           STA     [TMP_ACC]
           PUSHSR
           POPA
           STA     [TMP_SR]
           LDA     $AB
           STA     [TMP_IR]
           BSTX    [TMP_X]
           BSTSP   [TMP_SP]
           BSTIV   [TMP_IV]
           BLDX    $
           BSTX    [TMP_PC]
           JSR     [REGDUMP]
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function test_cmd: put any tests here
;
; Input:  --
;
; Output: --
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
TEST_CMD:  BLDX    TEST_CMD1
           JSR     [puts]
; test the keypad
           BLDX    TEST_CMD3
           JSR     [puts]
TEST_CMD5: JSR     [KEY_GETCH]
           JZ      [TEST_CMD6]
           PUSHA
           BLDX    TEST_CMD4
           JSR     [puts]
           POPA
           JSR     [PUTHEX]
           JNZ     [TEST_CMD5]
TEST_CMD6: JSR     [newline]
           RTS

TEST_CMD1: .BYTE   "This is function TEST_CMD", CR, LF, 0
TEST_CMD2: .BYTE   "The cmdline was:", CR, LF, 0
TEST_CMD3: .BYTE   "reading keypad ... (key '0' for exit)", CR, LF, 0
TEST_CMD4: .BYTE   CR, LF, "key was: $", 0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function loadram: load a RAM file
;
; Input:  --
;
; Output: 0 .. OK
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
LOADRAM:   LDA     0
           STA     [ERROR]
; read 4 comment lines at begin of file
           BLDX    BUF
           JSR     [gets]
           JSR     [gets]
           JSR     [gets]
           JSR     [gets]
; wait for start address
LOADRAM1:  JSR     [getch]
           CMPA    '$'
           JNZ     [LOADRAM1]
; read start address, high byte
           JSR     [LRAMBYTE]
           STA     [ADDR]
; low byte
           JSR     [LRAMBYTE]
           STA     [ADDR+1]

; wait for end address
LOADRAM2:  JSR     [getch]
           CMPA    '$'
           JNZ     [LOADRAM2]
; read start address, high byte
           JSR     [LRAMBYTE]
           STA     [END_ADDR]
; low byte
           JSR     [LRAMBYTE]
           STA     [END_ADDR+1]

; new line
LOADRAM3:  LDA     '.'            ; output a progress indicator
           JSR     [putchar]

; next character
LOADRAM4:  JSR     [getch]
           CMPA    CR
           JZ      [LOADRAM3]
           CMPA    LF
           JZ      [LOADRAM3]     ; next line
           CMPA    ' '
           JNZ     [LOADRAM4]     ; skip address at begin of line
           JSR     [LRAMBYTE]     ; read two characters and form a byte
           STA     [[ADDR]]
; increment address
           BLDX    [ADDR]
           INCX
           BSTX    [ADDR]
; check if end, high byte first
           LDA     [ADDR]
           CMPA    [END_ADDR]
           JC      [LOADRAM5]     ; HI(addr) > HI(end) => reached end
           JNZ     [LOADRAM4]     ; HI(addr) != HI(end) => next byte
; high bytes equal, compare low bytes
           LDA     [ADDR+1]
           CMPA    [END_ADDR+1]
           JNC     [LOADRAM4]     ; addr <= end => next byte

; we are done => eat up all trailing '--'
LOADRAM5:  JSR     [getch]
           CMPA    LF
           JZ      [LOADRAM6]
           CMPA    CR
           JNZ     [LOADRAM5]

LOADRAM6:  LDA     [ERROR]
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function lrambyte: receive two hex characters and form a byte
;
; Input:  --
;
; Output: byte in ACC
;
; Uses:   A, Z, TMP
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
LRAMBYTE:  JSR     [getch]        ; high nibble
           CMPA    '9'
           JC      [LRMBYTE1]
           SUB     '0'
           JMP     [LRMBYTE2]
LRMBYTE1:  OR      $20
           SUB     'a'-10
LRMBYTE2:  SHL
           SHL
           SHL
           SHL
           STA     [TMP]
           JSR     [getch]        ; low nibble
           CMPA    '9'
           JC      [LRMBYTE3]
           SUB     '0'
           JMP     [LRMBYTE4]
LRMBYTE3:  OR      $20
           SUB     'a'-10
LRMBYTE4:  OR      [TMP]
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function prompt: output a newline and a prompt to the RS-232
;
; Input:  --
;
; Output: --
;
; Uses:   A, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
PROMPT:    LDA     CR
           JSR     [putchar]
           LDA     LF
           JSR     [putchar]
           LDA     '#'
           JSR     [putchar]
           LDA     ' '
           JSR     [putchar]
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function newline: output a CR/LF pair to the RS-232
;
; Input:  --
;
; Output: --
;
; Uses:   A, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
newline:   LDA     CR
           JSR     [putchar]
           LDA     LF
           JSR     [putchar]
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function key_getch: read a character from the keypad
;
; Input:  --
;
; Output: the keycode in A
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
KEY_GETCH: LDA     [KEYPAD]
           JN      [KEY_GETCH]    ; if bit7 is set, no key available
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function LCDPUT: put a character to the LCD
;
; Input:  A .. the character
;
; Output: the character
;
; Uses:   A, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
LCDPUT:    STA     [MAINDISP]     ; put the character into the LCD
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function LCDPUTS: put a string to the LCD
; Input:  X .. address of string
;
; Output: number of characters written
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
LCDPUTS:   BSTX    [PTR16]        ; save address of string
           BLDX    0              ; load the index register with 0 
LCDPUTS1:  LDA     [[PTR16],X]    ; load accumulator with character
           JZ      [LCDPUTS2]     ; end of string => return
           STA     [MAINDISP]     ; write character to UART
           INCX
           JMP     [LCDPUTS1]     ; next character
LCDPUTS2:  BSTX    [PTR16]
           DECA
           LDA     [PTR16]        ; get number of written bytes
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function LCDPUTH: put a HEX byte to the LCD
;
; Input: A .. the digit
;
; Output: --
;
; Uses:  A, Z, Carry
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
LCDPUTH:   PUSHA
; process upper nibble
           SHR
           SHR
           SHR
           SHR
           AND     $0F
           CMPA    9
           JC      [LCDPUTH1]
           ADD     '0'
           JMP     [LCDPUTH2]
LCDPUTH1:  ADD     'A'
           SUB     10
LCDPUTH2:  STA     [MAINDISP]    ; write to LCD

; process lower nibble
           POPA
           AND     $0F
           CMPA    9
           JC      [LCDPUTH3]
           ADD     '0'
           JMP     [LCDPUTH4]
LCDPUTH3:  ADD     'A'
           SUB     10
LCDPUTH4:  STA     [MAINDISP]    ; write to LCD
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function atox: convert a string to an 8bit integer, interpret as HEX
;
; Input:  X .. ptr to string
;
; Output: A .. the 8bit number
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
ATOX:      LDA     2
           STA     [MAXDIGIT]
; call atolx
           JSR     [ATOLXX]
; adjust to 8bit
           BSTX    [TMP16]
           LDA     [TMP16+1]
           RTS
           
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function atolx: convert a string to a 16bit integer, interpret as HEX
;
; Input:  X .. ptr to string
;
; Output: X .. the 16bit number
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
ATOLX:     LDA     4
           STA     [MAXDIGIT]
ATOLXX:    BSTX    [PTR16]        ; save the ptr
           LDA     0
           STA     [NEGATIVE]
           INCA
           STA     [DIGITS]
           BLDX    0
           BSTX    [NUMBER]

; skip white sapce
ATOX1:     LDA     [[PTR16],X]
           CMPA    $08
           JNC     [ATOX2]        ; character < TAB => finished;
           CMPA    $0D
           JC      [ATOX3]        ; character > 0x0D => continue
ATOX8:     INCX
           JMP     [ATOX1]

; check for sign
ATOX3:     CMPA    ' '
           JZ      [ATOX8]        ; character == ' ' => continue
           CMPA    '+'
           JZ      [ATOX4]
           CMPA    '-'
           JNZ     [ATOX5]
; is negative
           LDA     1
           STA     [NEGATIVE]
ATOX4:     INCX
           LDA     [[PTR16],X]    ; skip '+', '-'

; detect illegal characters
ATOX5:     CMPA    '0'-1
           JNC     [ATOX2]        ; character too small => finished
           CMPA    '9'
           JC      [ATOX7]
; is numeric
           SUB     '0'
           JMP     [ATOX6]

; is alpha or else
ATOX7:     OR      $20            ; convert to lower case
           CMPA    'a'-1
           JNC     [ATOX2]        ; tolower(character) < 'a'
           CMPA    'f'
           JC      [ATOX2]        ; tolower(character) > 'f'
           SUB     'a'-10
; we have now the value in A
ATOX6:     PUSHA
; get what we have until now and shift left 4 bits, high byte first
           LDA     [NUMBER]
           SHL
           SHL
           SHL
           SHL
           STA     [NUMBER]
; high nibble of low byte to low nibble of high byte
           LDA     [NUMBER+1]
           SHR
           SHR
           SHR
           SHR
           AND     $0F
           OR      [NUMBER]
           STA     [NUMBER]
; low byte
           LDA     [NUMBER+1]
           SHL
           SHL
           SHL
           SHL
           STA     [NUMBER+1]
           POPA
           OR      [NUMBER+1]
           STA     [NUMBER+1]     ; store

           LDA     [DIGITS]
           INCA
           CMPA    [MAXDIGIT]
           JC      [ATOX2]        ; max. no of digits reached => finish
           STA     [DIGITS]
           JMP     [ATOX4]        ; next digit

ATOX2:     LDA     [NEGATIVE]
           JZ      [ATOX10]       ; not negative
; is negative
           LDA     0
           SUB     [NUMBER+1]
           STA     [NUMBER+1]
           LDA     0
           SUBC    [NUMBER]
           STA     [NUMBER]
ATOX10:    BLDX    [NUMBER]
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function strcpy: copy one string to another
;
; Input:  two 16bit pointers on the stack
;
; Output: X .. ptr to destination string
;
; Uses:   A, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
STRCPY:    POPA
           STA     [R_ADD]        ; save MSB of return address
           POPA
           STA     [R_ADD+1]      ; save LSB of return address

           POPA
           STA     [S2]           ; save LSB of src
           POPA
           STA     [S2+1]         ; save MSB of src
           POPA
           STA     [S1]           ; save LSB of dest
           POPA
           STA     [S1+1]         ; save MSB of dest
           BLDX    $FFFF

STRCPY1:   INCX
           LDA     [[S2], X]
           STA     [[S1], X]
           JZ      [STRCPY2]      ; end of src
           JMP     [STRCPY1]      ; next character

; src reached end
STRCPY2:   LDA     [R_ADD+1]
           PUSHA                  ; restore LSB of return address
           LDA     [R_ADD]
           PUSHA                  ; restore MSB of return address
           BLDX    [S1]           ; put result into X
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function strcmp: compare two null terminated strings
;
; Input:  two 16bit pointers on the stack
;
; Output: 0 ... s1 == s2
;         -1 .. s1 < s2
;         1 ... s1 > s2
;
; Uses:   A, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
STRCMP:    POPA
           STA     [R_ADD]        ; save MSB of return address
           POPA
           STA     [R_ADD+1]      ; save LSB of return address

           POPA
           STA     [S2]           ; save LSB of s2
           POPA
           STA     [S2+1]         ; save MSB of s2
           POPA
           STA     [S1]           ; save LSB of s1
           POPA
           STA     [S1+1]         ; save MSB of s1
           BLDX    $FFFF

STRCMP1:   INCX
           LDA     [[S2], X]
           JZ      [STRCMP2]      ; end of s2
           STA     [TMP]
           LDA     [[S1], X]
           JZ      [STRCMP3]      ; end of s1 but not of s2 => s2 > s1
           CMPA    [TMP]
           JZ      [STRCMP1]      ; equal => next character
           JC      [STRCMP4]      ; s1 > S2
STRCMP3:   LDA     $FF
           JMP     [STRCMP5]
STRCMP4:   LDA     1
           JMP     [STRCMP5]

; s2 reached end
STRCMP2:   LDA     [[S1], X]
           JNZ     [STRCMP4]      ; s1 not at end => s1 > s2
           LDA     0
STRCMP5:   STA     [TMP]
STRCMP6:   LDA     [R_ADD+1]
           PUSHA                  ; restore LSB of return address
           LDA     [R_ADD]
           PUSHA                  ; restore MSB of return address
           LDA     [TMP]          ; put result into acc
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function getarg: get the next argument from the cmdline
;
; Input:  2 16bit ptr to cmdline(src) and arg (dest) on stack
;
; Output: X .. ptr to arg
;         16bit ptr to next arg on stack
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
GETARG:
; save return address
           POPA
           STA     [R_ADD]
           POPA
           STA     [R_ADD+1]

           POPA
           STA     [S2]           ; save LSB of src
           POPA
           STA     [S2+1]         ; save MSB of src
           POPA
           STA     [S1]           ; save LSB of dest
           POPA
           STA     [S1+1]         ; save MSB of dest
           BLDX    $FFFF

GETARG1:   INCX
           LDA     [[S2], X]
           STA     [[S1], X]
           JZ      [GETARG2]      ; end of string
           CMPA    ' '
           JNZ     [GETARG1]      ; next character
; end of token, store terminating 0
           LDA     0
           STA     [[S1], X]
           INCX                   ; skip the ' '

; string reached end add X to input string and push onto stack
GETARG2:   BSTX    [TMP16]
           LDA     [TMP16+1]
           ADD     [S2+1]
           PUSHA
           LDA     [TMP16]
           ADDC    [S2]
           PUSHA
           
; restore return address
           LDA     [R_ADD+1]
           PUSHA
           LDA     [R_ADD]
           PUSHA
           BLDX    [S1]           ; put result into X
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function PUTHEX16: put a 16bit HEX number to the RS-232 (4 digits)
;
; Input: X .. the 16bit number
;
; Output: --
;
; Uses:  A, Z, Carry, TMP16
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
PUTHEX16:  BSTX    [TMP16]
           LDA     [TMP16]
           JSR     [PUTHEX]
           LDA     [TMP16+1]
           JSR     [PUTHEX]
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function PUTHEX: put a HEX byte to the RS-232 (2 digits)
;
; Input: A .. the byte
;
; Output: the byte in A
;
; Uses:  A, Z, Carry
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
PUTHEX:    PUSHA
           PUSHA
; process upper nibble
           SHR
           SHR
           SHR
           SHR
           AND     $0F
           CMPA    9
           JC      [PUTHEX1]
           ADD     '0'
           JMP     [PUTHEX2]
PUTHEX1:   ADD     'A'-10
PUTHEX2:   JSR     [putchar]

; process lower nibble
           POPA
           AND     $0F
           CMPA    9
           JC      [PUTHEX3]
           ADD     '0'
           JMP     [PUTHEX4]
PUTHEX3:   ADD     'A'-10
PUTHEX4:   JSR     [putchar]
           POPA
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function REGDUMP: display a register dump
;
; Input:  TMP_ACC, TMP_SR, TMP_X, TMP_SP, TMP_PC, TMP_IR
;
; Output: --
;
; Uses:   A, X
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
REGDUMP:
           BLDX    RD_MSG01
           JSR     [puts]
; program counter
           BLDX    RD_MSG02
           JSR     [puts]
           BLDX    [TMP_PC]
           JSR     [PUTHEX16]
           JSR     [newline]
; instruction register
           BLDX    RD_MSG03
           JSR     [puts]
           LDA     [TMP_IR]
           JSR     [PUTHEX]
           JSR     [newline]
; accumulator
           BLDX    RD_MSG04
           JSR     [puts]
           LDA     [TMP_ACC]
           JSR     [PUTHEX]
           JSR     [newline]
; status register
           BLDX    RD_MSG05
           JSR     [puts]
           LDA     [TMP_SR]
           JSR     [PUTHEX]
           JSR     [newline]
; index register
           BLDX    RD_MSG06
           JSR     [puts]
           BLDX    [TMP_X]
           JSR     [PUTHEX16]
           JSR     [newline]
; stack pointer
           BLDX    RD_MSG07
           JSR     [puts]
           BLDX    [TMP_SP]
           JSR     [PUTHEX16]
           JSR     [newline]

; interrupt vector
           BLDX    RD_MSG08
           JSR     [puts]
           BLDX    [TMP_IV]
           JSR     [PUTHEX16]
           JSR     [newline]

           RTS

RD_MSG01:  .BYTE   CR, LF, "DIY Calculator CPU register dump:", CR, LF, 0
RD_MSG02:  .BYTE   "  Program Counter      (PC): $", 0
RD_MSG03:  .BYTE   "  Instruction Register (IR): $", 0
RD_MSG04:  .BYTE   "  Accumulator          (A):  $", 0
RD_MSG05:  .BYTE   "  Status Register      (SR): $", 0
RD_MSG06:  .BYTE   "  Index Register       (X):  $", 0
RD_MSG07:  .BYTE   "  Stack Pointer        (SP): $", 0
RD_MSG08:  .BYTE   "  Interrupt Vector     (IV): $", 0

;===========================================================================
; global functions
;===========================================================================

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function PUTC: put a character to the RS-232
;
; Input:  A .. the character
;
; Output: the character
;
; Uses:   A, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
_PUTC:     PUSHA
PUTC1:     LDA     [TX_STOP]      ; CTRL-Q / CTRL-S
           JNZ     [PUTC1]
PUTC2:     LDA     [UART_STA]     ; read UART status
           AND     UART_TXE
           JZ      [PUTC2]        ; wait for transmitter empty
           POPA                   ; transmitter is empty now, get the char
           STA     [UART_TX]      ; ... and put it into the UART
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function PUTS: put a string to the RS-232
;
; Input:  X .. address of string
;
; Output: number of characters written
;
; Uses:   A, X, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
#if 1
_PUTS:     BSTX    [PTR16]        ; save address of string
           BLDX    0              ; load the index register with 0 
PUTS1:     LDA     [[PTR16],X]    ; load accumulator with character
           JZ      [PUTS2]        ; end of string => return
           JSR     [_PUTC]        ; write character to UART
           INCX
           JMP     [PUTS1]        ; next character
PUTS2:     BSTX    [PTR16]
           DECA
           LDA     [PTR16+1]      ; get number of transmitted bytes
           RTS
#else
_PUTS:     BSTX    [PTR16]        ; save address of string
           BLDX    0              ; load the index register with 0 
PUTS1:     LDA     [UART_STA]     ; read UART status
           AND     UART_TXE
           JZ      [PUTS1]        ; wait for transmitter empty
           LDA     [[PTR16],X]    ; load accumulator with character
           JZ      [PUTS2]        ; end of string => return
           STA     [UART_TX]      ; write character to UART
           INCX
           JMP     [PUTS1]        ; next character
PUTS2:     BSTX    [PTR16]
           DECA
           LDA     [PTR16]        ; get number of transmitted bytes
           RTS
#endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function TSTC: test if a charcter is available from the RS-232
;
; Input:  --
;
; Output: 0 .. no character available
;         1 .. character available
;
; Uses:   A, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
#if USE_RX_IRQ
_TSTC:     LDA     [RX_WPTR+1]
           CMPA    [RX_RPTR+1]
           JZ      [TSTC1]        ; no character available
           LDA     1
TSTC1:     RTS
#else
_TSTC:     LDA     [UART_STA]     ; read UART status
           AND     UART_RXR
           JZ      [TSTC1]        ; no byte ready
           LDA     1              ; load 1 into acc
TSTC1:     RTS
#endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function GETC: read a character from the RS-232
;
; Input:  --
;
; Output: the character
;
; Uses:   A, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
#if USE_RX_IRQ
_GETC:     LDA     [RX_WPTR+1]
           CMPA    [RX_RPTR+1]
           JZ      [_GETC]        ; no character available
           LDA     'G'
           JSR     [LCDPUT]
           LDA     [[RX_RPTR]]    ; read the byte
           PUSHA
; increment read pointer, low byte only
           LDA     [RX_RPTR+1]
           INCA
           STA     [RX_RPTR+1]
; check read pointer for wrap around, check only low byte
           JZ      [GETC1]        ; automatic wrap around
           CMPA    [RX_BMAX]
           JNC     [GETC1]        ; OK
           LDA     RX_BUF&$FF
           STA     [RX_RPTR+1]
; hardware handshake
           LDA     [RX_STOP]
           JZ      [GETC1]        ; was not stopped
           LDA     [RX_WPTR+1]
           CMPA    [RX_RPTR+1]
           LDA     [RX_RPTR+1]
           JNC     [GETC2]
           ADD     RX_BSIZE
GETC2:     SUB     [RX_WPTR+1]
           CMPA    RX_THRES
           JNC     [GETC1]        ; wptr-rptr <= threshold
; wptr - rptr > threshold => restart the device on the other end
           LDA     [UART_STA]
           OR      UART_RTS
           STA     [UART_CTL]
           LDA     0
           STA     [RX_STOP]
GETC1:     POPA
           RTS
#else
_GETC:
; set RTS line for HW handshake
           LDA     [UART_STA]
           JN      [GETC1]        ; RTS is in automatic mode
           OR      UART_RTS
           STA     [UART_CTL]
; wait for character
GETC1:     LDA     [UART_STA]     ; read UART status
           AND     UART_RXR
           JZ      [GETC1]        ; wait for byte available
           LDA     [UART_RX]      ; read the byte from the UART
           PUSHA
; reset RTS line
           LDA     [UART_STA]
           JN      [GETC2]        ; RTS is in automatic mode
           AND     ~UART_RTS
           STA     [UART_CTL]
GETC2:     POPA
           RTS
#endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function GETCHE: read a character from the RS-232 with echo
;
; Input:  --
;
; Output: the character
;
; Uses:   A, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
_GETCHE:    JSR     [_GETC]        ; get the character
            JSR     [_PUTC]        ; echo the character
            RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function GETCHAR: read a char from the RS-232 and wait for Enter
;
; Input:  --
;
; Output: the character
;
; Uses:   A, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
_GETCHAR:   JSR     [_GETC]        ; get a character
            JSR     [_PUTC]        ; echo it
            PUSHA                  ; save character
; wait for carriage return or line feed
GETCHAR1:   CMPA    CR
            JZ      [GETCHAR2]
            CMPA    LF
            JZ      [GETCHAR2]
            JSR     [_GETC]
            JMP     [GETCHAR1]

GETCHAR2:   POPA                   ; restore character
            RTS            

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function GETS: read a line from the RS-232
;
; Input:  X .. address of string
;
; Output: X .. address of string
;
; Uses:   A, Z
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
_GETS:     BSTX    [PTR16]        ; save address of string
           BLDX    0              ; load the index register with 0 
GETS1:     JSR     [_GETC]
; backspace support
           CMPA    BS
           JNZ     [GETS2]
           BSTX    [TMP16]
           LDA     [TMP16+1]
           JZ      [GETS1]        ; dont write over the left end
           LDA     BS             ; output a backspace
           JSR     [_PUTC]
           LDA     ' '            ; overwrite with ' '
           JSR     [_PUTC]
           DECX                   ; decrement index
           LDA     BS             ; output another backspace
           JSR     [_PUTC]
           JMP     [GETS1]        ; next character
; echo the character
GETS2:     JSR     [_PUTC]

; check for line end
           CMPA    CR
           JZ      [GETS3]        ; reached the end
           CMPA    LF
           JZ      [GETS3]        ; reached the end
           STA     [[PTR16],X]    ; save character
           INCX
           JMP     [GETS1]        ; next character

GETS3:     LDA     0
           STA     [[PTR16],X]    ; save the terminating '\0' character
; output a newline
           LDA     CR
           JSR     [_PUTC]
           LDA     LF
           JSR     [_PUTC]

           BLDX    [PTR16]        ; restore X
           RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function irq_hand: the main interrupt handler
;
; Input:  --
;
; Output: --
;
; Uses:
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
IRQ_HAND:
; save registers
;           PUSHA
;           PUSHX
           STA     [IRQ_ACC]
           BSTX    [IRQ_X]

; detect the interrupt source
           LDA     [IRQ_SRC]      ; must read only once!
           STA     [IRQ_ISRC]
           AND     IRQ_IRQ
           JNZ     [IRQ_H1]

           LDA     [IRQ_ISRC]
           AND     IRQ_UART
           JNZ     [IRQ_H2]

           LDA     [IRQ_ISRC]
           AND     IRQ_RUN
           JNZ     [IRQ_H3]

           LDA     [IRQ_ISRC]
           AND     IRQ_STEP
           JNZ     [IRQ_H4]
; software interrupt
           LDA     '&'
           JSR     [LCDPUT]
           JMP     [IRQ_HEND]

; external interrupt
IRQ_H1:    LDA     '*'
           STA     [MAINDISP]
           JMP     [IRQ_HEND]

; UART interrupt
IRQ_H2:
; detect if Rx or Tx interrupt
           LDA     [UART_STA]
           AND     UART_RXR
           JNZ     [RX_IRQ]       ; received a byte

           LDA     [UART_STA]
           AND     UART_TXE
           JNZ     [TX_IRQ]       ; transmitter free

           JMP     [IRQ_HEND]     ; unknown

; 'run' button interrupt
IRQ_H3:    LDA     '-'
           STA     [MAINDISP]
           JMP     [IRQ_HEND]

; 'step' button interrupt
IRQ_H4:    LDA     '~'
           STA     [MAINDISP]
           JMP     [IRQ_HEND]

; restore registers
IRQ_HEND:
;           POPX
;           POPA
           BLDX    [IRQ_X]
           LDA     [IRQ_ACC]
           RTI

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function rx_irq: the UART Rx data part of the interrupt handler
;
; Input:  --
;
; Output: --
;
; Uses:   RX_BUF, RX_WPTR
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
RX_IRQ:    LDA     'I'
           JSR     [LCDPUT]
           LDA     [UART_RX]      ; read byte from UART
           CMPA    CTRL_C
           JZ      [RX_IRQ2]
           CMPA    CTRL_S
           JZ      [RX_IRQ3]
           CMPA    CTRL_Q
           JZ      [RX_IRQ4]
           STA     [[RX_WPTR]]    ; store into ring buffer
; increment write pointer, low byte only
           LDA     [RX_WPTR+1]
           INCA
           STA     [RX_WPTR+1]
; check write pointer for wrap around, check only low byte
           JZ      [IRQ_HEND]     ; == 0 => automatic wrap around
           CMPA    [RX_BMAX]
           JC      [RX_IRQ1]
           LDA     '.'            ; OK
           JSR     [LCDPUT]
           JMP     [RX_IRQ5]

; reset write pointer to start of ring buffer
RX_IRQ1:   LDA     '+'
           JSR     [LCDPUT]
           LDA     RX_BUF&$FF
           STA     [RX_WPTR+1]
; hardware handshake
RX_IRQ5:   CMPA    [RX_RPTR+1]
           LDA     [RX_RPTR+1]
           JNC     [RX_IRQ6]
           ADD     RX_BSIZE
RX_IRQ6:   SUB     [RX_WPTR+1]
           CMPA    RX_THRES
           JC      [IRQ_HEND]     ; wptr-rptr > threshold
; wptr - rptr <= threshold => stop the device on the other end
           LDA     [UART_STA]
           AND     ~UART_RTS
           STA     [UART_CTL]
           LDA     1
           STA     [RX_STOP]
           JMP     [IRQ_HEND]     ; end of interrupt

; CTRL-C handler: jump to $0000
RX_IRQ2:   LDA     0
           PUSHA                  ; push $0000 as return addr onto the stack
           PUSHA
           PUSHA                  ; push $00 as status register onto the stack
           RTI                    ; a "legal" return from interrupt

; CTRL-S handler: stop sending via RS-232
RX_IRQ3:   LDA     1
           STA     [TX_STOP]
           JMP     [IRQ_HEND]

; CTRL-Q handler: restart sending via RS-232
RX_IRQ4:   LDA     0
           STA     [TX_STOP]
           JMP     [IRQ_HEND]

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function tx_irq: the UART Tx data part of the interrupt handler
;
; Input:  --
;
; Output: --
;
; Uses:   
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
TX_IRQ:
           JMP     [IRQ_HEND]     ; end of interrupt

           .END
